/**
*
* Orion 20m v2.0
* (c) 2011 AtomicDryad, orbiter-forums.com
* (c) 2004 Erik Anderson (Sputnik), erik_h_anderson@hotmail.com
*
*
* Licence LGPL -- http://www.gnu.org/licenses/lgpl.txt
*
* You may use this source code in your addon provided that if you
* publish your addon either privately or publicly the end user
* recieves a copy of the source code including any modifications
* you may have made. 
*
* 
*
* Requires CVEL library --                             
* please see ProjectGanymede on SourceForge for the latest source *
*/

#include "Orion20m.h"
#include "cvel.h"

ORION::ORION (OBJHANDLE hObj, int fmodel) : API (hObj) {
	this->cfg = new ORIONCFG();

	dprint("ORION: start. loadmesh: P...");
	hOrionP = oapiLoadMeshGlobal( "Orion/Orion20mP");
	dprint("ORION: loadmesh: M...");
	hOrionM = oapiLoadMeshGlobal( "Orion/Orion20mM");
	dprint("ORION: loadmesh: H...");
	hOrionH = oapiLoadMeshGlobal( "Orion/Orion20mH");
	dprint("ORION: loadmesh: done. RegisterExhaustTexture...");
	tex   = oapiRegisterExhaustTexture ("Exhaust");
	dprint("ORION: RegisterExhaustTexture: done.");

	MyID=ConnectToOrbiterSoundDLL3(hObj);
	havesound=true;
	RequestLoadVesselWave3(MyID,5,"Sound\\Orion\\atomicbomb.wav",DEFAULT);
	hudfont = oapiCreateFont (16, true, "Fixed", FONT_BOLD);	// fixme: cfg
	cfg->EmptyMass=EmptyMass;
	cfg->MainThrust=MainThrust;
	cfg->RCSThrust=RCSThrust;
	cfg->GyroThrust=GyroThrust;
	cfg->Ispmain=Ispmain;
	cfg->IspRCS=IspRCS;
	cfg->IspGYRO=IspGYRO;
	cfg->FuelMass=FuelMass;
	cfg->RCSFuelMass=RCSFuelMass;
	cfg->GyroFuelMass=GyroFuelMass;
	cfg->DestructRange=DestructRange;
	cfg->FlashSize=FlashSize;
	cfg->hudfontcolor=0xAAAA00;
	cfg->hudselcolor=0xFFFF33;
	cfg->hudshadowcolor=0x111111;
	cfg->hudmesgcolor=0xAAAAAA;
	cfg->hudfillcolor=0x555500;
	//// storable
	cfg->hudmode = HUD_MAIN;
	cfg->nukearmed = 1;
	cfg->nukesafety1 = 0;
	cfg->nukesafety2 = 0;
	cfg->nukesafety3 = 0;
	cfg->nukeyield=1.0;
	cfg->rcs_cog_comp=0;
	//// .cfg
	cfg->nukedanger = 1;
	// runtime
	cfg->nukewarned=0;
	//cfg->syslogp=0;
	cfg->syslogp=0;
	cfg->booting=1;
	//syslog("ORION v%d booting%s...",OrionVersion);//,(DEBUG>0?", debug: ON ":""));
	// todo: organize
	boomcount = 0;
	lcount=0;
	rot_comp_mode = 0;
	rot_comp_level = 0.0;
	sprintf(hudtitle,"[U]: Hud on.");
	rcs_shift_mode = 0;
	rcstimer=0;
	boomtimer=0;
	rcs_shift_vec=_V(0,0,0);
	shipcg=_V(0,0,0);
	SSCOUNT=0;
	ssmass=0.0;
#ifdef CVEL
	dprint("ORION: cvel...");
	CheckCentralMessage(); 
	ph_main=NULL;
#endif
	syslog("ORION v%f booting%s...",OrionVersion,(DEBUG>0?", debug: ON":""));
	dprint("ORION: done");
}

void ORION::clbkVisualCreated( VESSEL *vessel, VISHANDLE vis, int refcount ) {
	dprint("clbkVisualCreated: start");
#ifdef CVEL
	ClearMeshesEx();
	PlaceComponents();
    Draw(_V(0,0,0));
#endif
	dprint("clbkVisualCreated: done");
}

void ORION::SetMainStage() {
	dprint("SetMainStage: start");
#ifdef CVEL
	ClearMeshesEx();
	PlaceComponents();
    Draw(_V(0,0,0));
#endif
	dprint("SetMainStage: done");
}


void ORION::getsuperstructuremass(VESSEL * ship) {
	dprint("getsuperstructuremass: %s",ship->GetName());
	if (SSCOUNT>=MAXSUPERSTRUCTUREDEPTH) return;

	for (int x = 0; x < SSCOUNT; x++) {
		if(ship==superstructure[x].ref) {
			dprint("  skipping dupe: %s",ship->GetName());
			return;
		}
	}
	
	superstructure[SSCOUNT].ref=ship;
	SSCOUNT++;

	ssmass=ssmass + ship->GetMass();

	DOCKHANDLE dockH;
	OBJHANDLE objH;
	int docks=0;
	docks=ship->DockCount();
	dprint("  [%s] docks: %d",ship->GetName(),docks);
	for (int i = 0; i < docks; i++) {
		if(ship->DockingStatus(i)) {
			dockH=ship->GetDockHandle(i);
			objH=ship->GetDockStatus(dockH);
			VESSEL * veh = oapiGetVesselInterface(objH);
			dprint("  [%s] dock %d/%d: %s",ship->GetName(),i,docks,veh->GetName());
			getsuperstructuremass(veh);
		}
	}
	return;
}

void ORION::clbkPreStep(double simt, double simdt, double mjd) {
	VECTOR3 cg;
	static double rcst=0.0;
	rcst += oapiGetSysStep();
	
	if (rcst>1.000) {		
		rcst=0.0;
		if(cfg->rcs_cog_comp==1) {
			GetSuperstructureCG(cg);
			double cgdist=dist(cg,shipcg);
			if(cgdist>0.0) {
				dprint("(%s)newcg: %.8f, (%.4f,%.4f,%.4f) -> (%.4f,%.4f,%.4f)",GetName(),cgdist,shipcg,cg);
				//shipcg=cg;
				rcs_linshift(2);
			}
		}
		rcst=0.0;
	}
}

void ORION::clbkPostStep(double simt, double simdt, double mjd) {

#ifdef CVEL
	if( CheckCentralMessage() ) SetMainStage();
#endif
	da = oapiGetSimStep() * SLAM_SPEED;
	SetThrusterGroupLevel(thg_orion, 0.0);

	SetAnimation (anim_slam, slam_proc);

	boom->Activate(0);
// Engine is on when the main throttle is up, or to continue the explosion cycle
	if (((GetThrusterGroupLevel(thg_main)) > 0.25) || (slam_proc-da > 0.0))
		SlamCycle(da);
	if (slam_proc-da<0.0)			// if we're NOT cycling, re-vector the thrust angle
		VectorThrust();

	ummu_clbkPostStep(simt,simdt,mjd);
/*
	  if(gpds_gui.OperationComplete() && gpds_gui_active )
		{
		    // set this flag false to make sure
		    // we dont execute this routine again until
		    // we have run the GUI
			gpds_gui_active=false;

			// clear for a redraw
			ClearMeshesEx();

			// read in the created GPDS file
			ReadGPDSFile( "gpdsiofile.dat");
			PlaceComponents();
			Draw();
		}
		*/

}


void ORION::SlamCycle(double da)
{
		if(!cfg->nukearmed) {
			if (cfg->nukewarned != 1) {
				cfg->nukewarned=1;
				syslog("ERROR: drive is OFFLINE. [ALT-A] to arm nukes.");
			}
			return;
		}

		PARTICLESTREAMSPEC exhaust_main = {
			0, 300, 0.4, 500, 0, 12, 50, 0.0, PARTICLESTREAMSPEC::EMISSIVE,
			PARTICLESTREAMSPEC::LVL_SQRT, 0, 1,
			PARTICLESTREAMSPEC::ATM_PLOG, 0, 0.1
		};
		if((havesound) && (firstnukestep) && (GetHandle() == oapiGetFocusObject())) {
			StopVesselWave3(MyID,5);
			PlayVesselWave3(MyID,5);								// Play the sound!  BIG boom!
			cloud = AddExhaustStream (th_orion[0],ExplosionPosition, &exhaust_main);
		}
		if (firstnukestep) { 
			
			
			firstnukestep = false;
			DestroyNeighbors(true);
		// todo: remove bomb from fuel at eject
			boomcount++;
#ifdef BOMBFUEL
			double vehfuelmass=GetFuelMass();
			SetFuelMass(vehfuelmass-500);
#endif
			
		}
		if (!MoveNegative) slam_proc = slam_proc + da*2;	// Plate moves faster compressing...
			else slam_proc = slam_proc - da;				// ...than expanding

		if (slam_proc+da*2 > 1.0) MoveNegative = true;		// Now we're expanding
		if (slam_proc-da < 0.0) 
		{
			slam_proc=0.0;				// FORCE the cycle ended
			MoveNegative = false;		// Now we'll compress, next bomb
		};
		SetThrusterGroupLevel(thg_orion, 1.0*cfg->nukeyield);				// Set Orion thrust "on"

// Vary the exhaust texture to make the flash
		DelExhaust(nuke);
		boom->Activate(0);
		nuke = AddExhaust (th_orion[0], 1.0, cfg->FlashSize*slam_proc*slam_proc*slam_proc*slam_proc, tex);
		boom->Activate(1);
		boom->SetIntensity((slam_proc*2)+0.5);

		if ((MoveNegative) || (slam_proc > da)) {
			DelExhaustStream (cloud);
		}// Stop making smoke puffs once we're done booming
/*
// Jiggle the ship a bit
		bombcounter+=1;
		if (bombcounter > 3) bombcounter=0;
		tempP = 500000*sin(fabs(jiggle_proc)*2*PI);
		tempY = 500000*sin(fabs(jiggle_proc)*2*PI);
		if (bombcounter == 1) tempP=tempP*-1;
		if (bombcounter == 2) tempY=tempY*-1;
		if (bombcounter == 3) {
			tempP=tempP*-1;
			tempY=tempY*-1;
		}
		AddForce(_V(tempY, tempP, 0), _V(0,0,20));
		AddForce(_V(-tempY, -tempP, 0), _V(0,0,-20));
		jiggle_proc+=(da*1.5);
		if (jiggle_proc > 1.0) jiggle_proc = -1.0;
*/
}

void ORION::VectorThrust()
{
// Vector the thrust (exploding bombs slightly off-center) 
// Only happens BETWEEN explosions, of course, never during them.
		
		if (oapiGetTimeAcceleration() > 10) return;
	
		tempP = GetManualControlLevel(THGROUP_ATT_PITCHDOWN, MANCTRL_ANYDEVICE, MANCTRL_ANYMODE) - 
			GetManualControlLevel(THGROUP_ATT_PITCHUP, MANCTRL_ANYDEVICE, MANCTRL_ANYMODE);

		tempY = GetManualControlLevel(THGROUP_ATT_YAWLEFT, MANCTRL_ANYDEVICE, MANCTRL_ANYMODE) - 
			GetManualControlLevel(THGROUP_ATT_YAWRIGHT, MANCTRL_ANYDEVICE, MANCTRL_ANYMODE);
	 
		GetAngularVel(AngularVel);
		tempP+=(10*AngularVel.x);			// Bias the thrust vector AGAINST the angular velocity a bit
		tempY+=(-10*AngularVel.y);			// This will damp angular velocity, preventing loss of control

		
		double fuelmass=GetFuelMass();

		double vlevel=((0.007*fuelmass)/cfg->FuelMass)+0.001; // crazy jiggling with low fuel

		SetThrusterDir(th_orion[0],_V(vlevel*tempY,vlevel*tempP,1));
		firstnukestep = true;
}

void ORION::PlaceComponents()
{
	dprint("PlaceComponents: start");
	VECTOR3 pos;
#ifdef CVEL
	COMMON_MESH tmp;

	pos = _V(0,0,-13.0);
	tmp.MeshHandle(hOrionP);			// Propulsion module
	tmp.MeshPosition( pos);				// Pusher, struts, and many of the bombs
	AddMeshEx( tmp);
	pos = _V( 0,0,7.5 );
	tmp.MeshHandle(hOrionM);			// Magazine (fuel tanks, sorta)
	tmp.MeshPosition( pos);				// Note:  Magazine also includes some storage
	AddMeshEx( tmp);					//		for mission payload
	pos = _V(0,0,29.0);
	tmp.MeshHandle(hOrionH);			// Hab area
	tmp.MeshPosition( pos);				// Hab, docking areas, and much else
	AddMeshEx( tmp);
#else
	pos = _V(0,0,-13.0);
	AddMesh(hOrionP,&pos);
	pos = _V( 0,0,7.5 );
	AddMesh(hOrionM,&pos);
	pos = _V(0,0,29.0);
	AddMesh(hOrionH,&pos);
#endif
	dprint("PlaceComponents: done");
}

void ORION::DestroyNeighbors(bool nukestep) {
	if(nukestep!=true) return; // wtf bleh this procs twice...simstep is wonky...fixlater.
	if(cfg->nukedanger==0) return;

	sprintf(oapiDebugString(), "");
	Local2Global (ExplosionPosition, gorion);  // global position of the nuke explosion
	j = oapiGetVesselCount();
	for (i = 0; i < j; i++) {
		j = oapiGetVesselCount();			// Last-ditch check to prevent using oapiGetVesselByIndex
		if (i >= j) return;					// if some other vessel has changed the index on us
		hV = oapiGetVesselByIndex (i);
		if (hV == GetHandle()) continue; // we don't want to destroy ourselves ...
		oapiGetGlobalPos (hV, &gpos);
		distance = dist(gpos,gorion); // 10,10,10 11,11,11 = 1.73205
		if (distance < cfg->DestructRange) { // in range
			Global2Local (gpos,lpos); // 1,1,1 ///// 1,1,-76
			olpos=lpos;
			lpos.x=lpos.x/distance; // .57735
			lpos.y=lpos.y/distance; // .57735
			lpos.z=lpos.z+-(ExplosionPosition.z/2); // -76+(-75/2) = -113.5
			lpos.z=lpos.z/distance; // -113.5/1.73205 = -65.52
			if ((lpos.z < 0.0) || (lpos.x*lpos.x+lpos.y*lpos.y > lpos.z*lpos.z))  // Small angle-off from forward axis
				// .333333+.333333 = .66666 > -131.04 //////
													// We won't hit anything protected by our pusher plate
			{
				if (hV == oapiGetFocusObject())			// If we're trying to destroy the focus object
					oapiSetFocusObject(GetHandle());	// Switch to the destroying object
					
				VESSEL *remotevessel = oapiGetVesselInterface(hV);					
				VESSELSTATUS hv;
				oapiGetObjectName(hV,destructname,256);	
				dprint("Destroy: %s: dist: %.2f, gpos(%.2f,%.2f,%.2f) lpos(%.2f,%.2f,%.2f -> %.2f,%.2f,%.2f)",destructname,distance,gpos,olpos,lpos);
/*
					oapiGetGlobalPos (GetHandle(), &gpos);
					remotevessel->Global2Local (gpos,lpos);		// Check to see if it's protected
					lpos.x=lpos.x/distance;						// by its pusher plate
					lpos.y=lpos.y/distance;
					lpos.z=lpos.z/distance;
*/
				if (!_strnicmp (remotevessel->GetClassName(), "Orion", 5) != 0)		// If the target is an Orion also,
				{				
					oapiGetGlobalPos (GetHandle(), &gpos);
					remotevessel->Global2Local (gpos,lpos);		// Check to see if it's protected
					lpos.x=lpos.x/distance;						// by its pusher plate
					lpos.y=lpos.y/distance;
					lpos.z=lpos.z/distance;
					if ((lpos.z < 0.0) && (lpos.x*lpos.x+lpos.y*lpos.y < lpos.z*lpos.z)) { // Small angle-off from forward axis	
						sprintf(oapiDebugString(),"%s protected by pusher plate",destructname);
						continue;
					}
				}
	
				
				if ((remotevessel->GetFuelMass() > 0.0) && (distance > (cfg->DestructRange / 2))) {					
					remotevessel->SetFuelMass(0.0);		// Now it's a fuel-less, non-maneuvering wreck
					sprintf(oapiDebugString(),"Damaged %s due to nuclear blast",destructname);
					syslog("Damaged %s due to nuclear blast",destructname);
				}				
				else {
					remotevessel->Undock(ALLDOCKS);		// Undock everything.  Doesn't work right otherwise
					remotevessel->SetFuelMass(0.0);		// Damage it too.  Wouldn't want to forget that
					remotevessel->GetStatus(hv);
//					oapiDeleteVessel(hV);				// This works, but screws up the index
					OBJHANDLE oh=remotevessel->GetGravityRef();
					int wtf=oapiGetObjectType(oh);
					if(wtf!=4) { // != 4 == not a planet == crash
						remotevessel->AddForce(operator * (lpos,500000000),lpos); //todo: this could be better with spin and such
						//syslog("DESTROYED %s due to nuclear blast (%i) %.0f,%.0f,%.0f /  %.0f,%.0f,%.0f",destructname,wtf,lpos,operator * (lpos,500000000));
					} else {
						hv.status=1;						// Instead, we'll set status to "landed"
						remotevessel->DefSetState(&hv);		// That'll fix 'em!
					}
					sprintf(oapiDebugString(),"Destroyed %s due to nuclear blast",destructname);
					syslog("DESTROYED %s due to nuclear blast.",destructname);
				}

			}
		}
	}
}
#define getconfdouble(d,s) {\
	if ( (oapiReadItem_float(CFG,s,dvar)) && dvar>0) {\
		cfg->d=dvar;\
		if(logents<=6) {\
			sprintf_s(conflogstr,1024,"%s %s=%.8g",conflogstr,s,cfg->d);logents++;\
		} else {\
			syslog(conflogstr);\
			sprintf_s(conflogstr,1024,"  %s=%.8g",s,cfg->d);\
			logents=1;\
		}\
	}\
}

#define getconfvecD(d,s) {\
	if ( (oapiReadItem_vec(CFG,s,vecvar)) ) {\
		cfg->d=oapiGetColour((DWORD)vecvar.x,(DWORD)vecvar.y,(DWORD)vecvar.z);\
		if(logents<=6) {\
			sprintf_s(conflogstr,1024,"%s %s=%.0f,%.0f,%.0f",conflogstr,s,vecvar);logents++;\
		} else {\
			syslog(conflogstr);\
			sprintf_s(conflogstr,1024,"  %s=%.0f,%.0f,%.0f",s,vecvar);\
			logents=1;\
		}\
	}\
}

void ORION::clbkSetClassCaps(  FILEHANDLE CFG )
{
	//syslog("/etc/rc.S: [OK]");
	dprint("clbkSetClassCaps: start");
	int ivar,logents;double dvar;VECTOR3 vecvar;
	logents=0;
	char conflogstr[1024];
	sprintf(conflogstr,"/etc/rc.S: ");
	getconfdouble(EmptyMass,"EmptyMass");
	getconfdouble(MainThrust,"MainThrust");
	getconfdouble(RCSThrust,"RCSThrust");
	getconfdouble(GyroThrust,"GyroThrust");
	getconfdouble(Ispmain,"Ispmain");
	getconfdouble(IspRCS,"IspRCS");
	getconfdouble(IspGYRO,"IspGYRO");
	getconfdouble(GyroFuelMass,"GyroFuelMass");
	getconfdouble(RCSFuelMass,"RCSFuelMass");
	getconfdouble(FuelMass,"FuelMass");
	getconfdouble(DestructRange,"DestructRange");
	getconfdouble(FlashSize,"FlashSize");
	getconfvecD(hudfontcolor,"HudFontColor");
	getconfvecD(hudshadowcolor,"HudShadowColor");
	getconfvecD(hudselcolor,"HudSelColor");
	getconfvecD(hudmesgcolor,"HudMesgColor");
	getconfvecD(hudfillcolor,"HudFillColor");
	if(logents>0) syslog(conflogstr);
	SetUpOrion();
	syslog("/etc/rc.d/rc.M: [OK] - Vessel control online.",GetEmptyMass());
	clbkSetClassCaps_UMMu();
	SelectedUmmuMember=0;

	dprint("clbkSetClassCaps: done");
}

void ORION::SetUpOrion()
{
	dprint("SetUpOrion: start");
	ph_main=NULL;
	slam_proc = 0.0;
//	jiggle_proc = -1.0;
//	bombcounter = 0;
	numevas = 0;
	firstnukestep = true;

	dock0 = CreateDock (_V( 0.0, 0.0 , 42.0), _V( 0, 0, 1), _V(0,1, 0));
	dock1 = CreateDock (_V( 3.6, 0.0 , 33.575), _V( 1, 0, 0), _V(0,0,-1));
	dock2 = CreateDock (_V(-3.6, 0.0 , 33.575), _V(-1, 0, 0), _V(0,0,-1));
	dock3 = CreateDock (_V( 0.0, 3.8 , 26.6), _V(0,  1, 0), _V(0,0, 1));
	dock4 = CreateDock (_V( 0.0,-3.8 , 26.6), _V(0, -1, 0), _V(0,0, 1));
	dock5 = CreateDock (_V( 3.6, 0.0 , 23.9), _V( 1, 0, 0), _V(0,0, 1));
	dock6 = CreateDock (_V(-3.6, 0.0 , 23.9), _V(-1, 0, 0), _V(0,0, 1));
	dock7 = CreateDock (_V( 0.0, 6.5 , 18.2), _V( 0, 0, 1), _V(0,1, 0));
	dock8 = CreateDock (_V( 0.0, -6.50 , 18.2), _V(0, 0, 1), _V(0,-1, 0));


	SetSize (45);
	SetTouchdownPoints (_V(0,8.0,-24), _V(6,-8.0,-24), _V(-6,-8.0,-24));
	SetPMI (_V(134,117,52));
	SetCrossSections (_V(84,115,35));
	SetCW (0.13, 0.5, 1.4, 1.4);
	SetRotDrag (_V(1.5,1.0,1.0));
	SetPitchMomentScale (1e-6);
	SetBankMomentScale (1e-6);
	ph_main = CreatePropellantResource( cfg->FuelMass );
	SetDefaultPropellantResource( ph_main );

	ph_RCS = CreatePropellantResource(cfg->RCSFuelMass);
	ph_GYRO = CreatePropellantResource(cfg->GyroFuelMass);
#ifdef CVEL
	SetEmptyMass ( cfg->EmptyMass + CalculateMass () );
//	SetEmptyMass ( CalculateMass (EmptyMass));
	MoveNegative=true;
//	AddMesh( hOrion, &pos );
	ClearMeshesEx();
	PlaceComponents();
    Draw(_V(0,0,0));
#endif
	//SetCameraOffset(_V(0,0,41.0));
	SetCameraOffset(_V(0,0,38.0));
	SetOrionStage();
	EnableTransponder(true);
	UnDockArm = false;
//	syslog("init: done.");
	dprint("SetUpOrion: done. em=%f",GetEmptyMass());
}


void ORION::clbkSaveState( FILEHANDLE scn )
{
	dprint("clbkSaveState: start")
	SaveDefaultState (scn);

	/* CVE-lite function ! */
	/* IMPORTANT */
	/* Call this after the normal SaveDefautState() to record attached payloads
	/* it the scenario file */

	/* save payload lines */
#ifdef CVEL
	SaveDefaultStateEx(scn);
#endif
	///todo
	Crew.SaveAllMembersInOrbiterScenarios(scn);
	oapiWriteScenario_int (scn, "HUDMODE",cfg->hudmode);
	oapiWriteScenario_int (scn, "EVASEL",SelectedUmmuMember);
	oapiWriteScenario_int (scn, "DOCKSEL",SelectedUmmuDock);
	oapiWriteScenario_int (scn, "DOCKLIGHT",evalight->IsActive());
	oapiWriteScenario_int (scn, "COG_COMP",cfg->rcs_cog_comp);
	oapiWriteScenario_float(scn,"NUKEYIELD",cfg->nukeyield);
	oapiWriteScenario_int(scn,"NUKEARMED",cfg->nukearmed);
	char tmpstr[6];
	sprintf(tmpstr,"%1i %1i %1i",cfg->nukesafety1,cfg->nukesafety2,cfg->nukesafety3);
	tmpstr[6]=0;
	oapiWriteScenario_string(scn,"NUKESAFE",tmpstr);
	//oapiWriteScenario_int(scn,"NUKELOCK",cfg->nukesafety1+(cfg->nukesafety2*10)+(cfg->nukesafety2*100));
	dprint("clbkSaveState: done")
}

void ORION::clbkLoadStateEx(FILEHANDLE scn, void *vs)
{
	cfg->booting=1;
	dprint("clbkLoadStateEx: start");
    char *line;
	int ivar,ivar2,ivar3;
	float fvar;
	while (oapiReadScenario_nextline (scn, line)) 
	{
		/* CVE-lite function ! */
		/* IMPORTANT */
		/* Call this for every line in the scenario file. It will automatically
		/* deduce which lines are for it, and load the payloads from the file. */

		/* load payload lines */
#ifdef CVEL
		ParsePayloadString(line);
#endif

		ParseScenarioLineEx (line, vs);
		if (!_strnicmp (line, "HUDMODE", 7)) {
			if(sscanf (line +7, "%d", &ivar)) {
				dprint("got HUDMODE: %d",ivar);
				chhud(ivar);
			}
			continue;
		} else if (!_strnicmp (line, "EVASEL", 6)) {
			if(sscanf (line +6, "%d", &ivar)) {
				dprint("got EVASEL: %d",ivar);
				ummu_selectmember(ivar);
			}
			continue;
		} else if (!_strnicmp (line, "DOCKSEL", 7)) {
			if(sscanf (line +7, "%d", &ivar)) {
				dprint("got DOCKSEL: %d",ivar);
				ummu_selectdock(ivar);
			}
			continue;
		} else if (!_strnicmp (line, "DOCKLIGHT", 9)) {
			if(sscanf (line +9, "%d", &ivar)) {
				dprint("got DOCKLIGHT: %d",ivar);
				if(ivar==1) docklight();
			}
			continue;
		} else if (!_strnicmp (line, "COG_COMP", 8)) {
			if(sscanf (line +8, "%d", &ivar)) {
				dprint("got COG_COMP: %d",ivar);
				cfg->rcs_cog_comp=ivar;
			}
			continue;
		} else if (!_strnicmp (line, "NUKEYIELD", 9)) {
			if(sscanf (line +9, "%f", &fvar)) {
				dprint("got NUKEYIELD: %f",fvar);
				if(fvar >= 0.10 && fvar <= 1.0) cfg->nukeyield=fvar;
			}
			continue;
		} else if (!_strnicmp (line, "NUKEARMED", 9)) {
			if(sscanf (line +9, "%d", &ivar)) {
				dprint("got NUKEARMED: %d",ivar);
				cfg->nukearmed=ivar;
			}
			continue;
		} else if (!_strnicmp (line, "NUKESAFE", 8)) {
			if(sscanf (line +8, "%d %d %d", &ivar,&ivar2,&ivar3)) {
				dprint("got NUKESAFE: %d %d %d",ivar,ivar2,ivar3);
				cfg->nukesafety1=ivar;
				cfg->nukesafety2=ivar2;
				cfg->nukesafety3=ivar3;
			}
			continue;
		}
		if(Crew.LoadAllMembersFromOrbiterScenario(line)==TRUE) continue;
	}
#ifdef CVEL
	SetEmptyMass ( cfg->EmptyMass + CalculateMass () );
#endif
	/*
	int ivar;
	if(oapiReadItem_int(scn, "HUDMODE", ivar)) {
		dprint("got HUDMODE: %d",ivar);
	}
	*/
	cfg->booting=0;
	syslog("/etc/rc.d/rc.local: [OK] - init: done.");
	dprint("clbkLoadStateEx: done");
}

void ORION::SetOrionStage( )
{
	dprint("SetOrionStage: start");

	PARTICLESTREAMSPEC contrail = 
	{
		0, 5.0, 5, 250, 0.3, 4.0, 16, 3.0, PARTICLESTREAMSPEC::DIFFUSE,
		PARTICLESTREAMSPEC::LVL_SQRT, 0, 1,
		PARTICLESTREAMSPEC::ATM_PLOG, 1e-4, 1
	};

	PARTICLESTREAMSPEC exhaust_main = {
		0, 300, 0.5, 500, 0, 7, 50, 0.0, PARTICLESTREAMSPEC::EMISSIVE,
		PARTICLESTREAMSPEC::LVL_SQRT, 0, 1,
		PARTICLESTREAMSPEC::ATM_PLOG, 0, 0.1
	};
	exhaust_main.tex = oapiRegisterParticleTexture ("Contrail2");

/*	CreateAirfoil (LIFT_VERTICAL,   _V(0,0,-0.5), VLiftCoeff, 20, 270, 2.266);
	CreateAirfoil (LIFT_HORIZONTAL, _V(0,0,-4), HLiftCoeff, 20,  50, 1.5);

	CreateControlSurface (AIRCTRL_ELEVATOR, 5.0, 1.5, _V( 0, 0,  -15), AIRCTRL_AXIS_XPOS);
	CreateControlSurface (AIRCTRL_RUDDER,   2.0, 1.5, _V( 0, 3,  -16), AIRCTRL_AXIS_YPOS);
	CreateControlSurface (AIRCTRL_AILERON,  3.0, 1.5, _V( 7,-0.5,-15), AIRCTRL_AXIS_XPOS);
	CreateControlSurface (AIRCTRL_AILERON,  3.0, 1.5, _V(-7,-0.5,-15), AIRCTRL_AXIS_XNEG);

	PARTICLESTREAMSPEC rps = {
		0, 20, 20, 0, 0.03, 0.5, 100, 3, PARTICLESTREAMSPEC::DIFFUSE,
		PARTICLESTREAMSPEC::LVL_FLAT, 1, 1, PARTICLESTREAMSPEC::ATM_PLIN, 6e7, 12e7
	};
	AddReentryStream (&rps);
*/

	static UINT SlamGrp[5] = {1, 2, 3, 6, 8};		// 1, 2, 3, 5, 6, 8.  #9 is highest
	static MGROUP_TRANSLATE Slam (0, SlamGrp, 5,
		_V(0.0,0.0, 10.5));
	anim_slam = CreateAnimation (1.0);
	AddAnimationComponent (anim_slam, 0.0, 1.0, &Slam);
	SetAnimation (anim_slam, slam_proc);


	th_main[0] = CreateThruster (ExplosionPosition,    _V( 0.0,0.0,1.0), 
		1.0, ph_main, cfg->Ispmain, cfg->Ispmain);
	th_orion[0] = CreateThruster (ExplosionPosition,    _V( 0.0,0.0,1.0), 
		cfg->MainThrust, ph_main, cfg->Ispmain, cfg->Ispmain);

	thg_main  = CreateThrusterGroup( th_main, 1, THGROUP_MAIN );
	thg_orion = CreateThrusterGroup( th_orion, 1, THGROUP_USER );

//	cloud = AddExhaustStream (th_orion[0],ExplosionPosition, &exhaust_main);
	SoundOptionOnOff3(MyID,PLAYMAINTHRUST,FALSE);
	SoundOptionOnOff3(MyID,PLAYATTITUDETHRUST,FALSE);

	nuke = AddExhaust (th_orion[0], 1.0, 100.0, tex);
	SetThrusterGroupLevel(thg_orion, 0.0);
	COLOUR4 col_d = {0.9,0.9,0.8,0};
	COLOUR4 col_s = {0.9,0.9,0.8,0};
	COLOUR4 col_a = {0,0,0,0};
	COLOUR4 col_white = {1,1,1,0};
	boom= AddPointLight (_V(0,0,-200), 5000, 1e-4, 0, 2e-4, col_d, col_s, col_a);
	boom->SetIntensity(2);
	boom->Activate(0);
	dprint("SetOrionStage: SetAttControlsOrion");
	SetAttControlsOrion();
	EnableTransponder(true);
	dprint("SetOrionStage: done");
}


void ORION::chhud(int num) {
	char *menuTitle[HUDMENUCOUNT] = {"Hud off","Main", "RCS", "MMU/Docking"};//,"Lighting"}; // FIXME: dynamic
	if(num>HUDMENUCOUNT) return;
	cfg->hudmode=num;
	int nextmode=num+1;
	if(nextmode>=HUDMENUCOUNT) nextmode=0;
	sprintf(hudtitle,"%s menu -- [U] %s",menuTitle[num],menuTitle[nextmode]);
	/*
	for (int i = 0; i < 3; i++) {
		if(i+1!=num) {
			//sprintf(buf," [%d] %s menu",i+1,menuTitle[i]);
			sprintf(buf," %s",menuTitle[i]);
			strcat(hudtitle,buf);
		}
	}*/
}

void ORION::chyield(int updown) {
	double newyield;
	if(!updown) {
		newyield=cfg->nukeyield/2;
		if(newyield<0.25) newyield=0.10;
	} else {
		newyield=cfg->nukeyield*2;
		if(newyield<0.25) newyield=0.25;
		else if(newyield>1.00) newyield=1.00;
	}
	cfg->nukeyield=newyield;
	syslog("Bomb yield set to %.0f%%",newyield*100);
}

int ORION::armnuke(int onoff) {
	if(onoff==cfg->nukearmed) {
		syslog("ERROR: Nuclear pulse engine is already %s.",(onoff?"online":"offline"));
	} else if(onoff) {
		if(cfg->nukesafety1+cfg->nukesafety2+cfg->nukesafety3 > 0) {
			syslog("ERROR: Nuclear safeties set. To arm, enter: %s%s%s.",
				(cfg->nukesafety1?" [ALT-1]":""),
				(cfg->nukesafety2?" [ALT-2]":""),
				(cfg->nukesafety3?" [ALT-3]":"")
			);
		} else {
			syslog("WARNING: Nuclear pulse engine now ONLINE.");// Yield: %.0f%",cfg->nukeyield*100);
			cfg->nukearmed=1;
		}
	} else {
		syslog("Nuclear pulse engine now OFFLINE.");
		cfg->nukearmed=0;
		cfg->nukewarned=0;
	}
	return cfg->nukearmed;
}






/*
int ORION::clbkConsumeBufferedKey(DWORD key, bool down, char *keystate)
{


		if (!down) return 0; // only process keydown events

		if (KEYDOWN (keystate, OAPI_KEY_J) && !KEYMOD_CONTROL(keystate) && !KEYMOD_SHIFT(keystate) ) 
		{ // "Jettison Stage"
			if (oapiAcceptDelayedKey (OAPI_KEY_J, 1.0))
			{
				VESSEL *depV;
				VECTOR3 glopos, locpos;
				VESSELSTATUS vs;
				double payrot;

				depV = DeployPayload(_V(0,0,0));
				if (depV || NULL) {
					depV->GetGlobalPos(glopos);
					Global2Local (glopos, locpos);
					depV->GetStatus(vs);
					if ((strnicmp (depV->GetClassName(), "zrot", 4) == 0)
						|| (strnicmp (depV->GetClassName(), "Orion/zrot", 10) == 0)) {	// If the target class starts with "zrot" in it, 
																				//	then we can do this routine
						payrot = atan(locpos.x/locpos.y);
						if ((locpos.y < 0)) payrot += PI;
						if ((strnicmp (depV->GetClassName(), "zrot180", 7) != 0)
							&& (strnicmp (depV->GetClassName(), "Orion/zrot180", 13) != 0)) // If it starts with zrot180, spin it another 180
							payrot += PI;
						vs.arot.z += payrot;
						depV->DefSetState(&vs);
					}
				}

				ClearMeshesEx();
				PlaceComponents();
				Draw(_V(0,0,0));
				SetEmptyMass ( cfg->EmptyMass + CalculateMass () );
				SetAttControlsOrion();
			}
			UnDockArm = false;
			return 1;
		}
		if (KEYDOWN (keystate, OAPI_KEY_C) && KEYMOD_CONTROL(keystate) && !KEYMOD_SHIFT(keystate) ) 
		{ // "Jettison Stage"
//			if (oapiAcceptDelayedKey (OAPI_KEY_C, 1.0))
//				dock1=CreateDock( _V(0,0.0,2.0), _V(0,0,1), _V(0,1,0));
			
			UnDockArm = false;
			return 1;
		}
		

		if (KEYDOWN (keystate, OAPI_KEY_E) && !KEYMOD_CONTROL(keystate) && !KEYMOD_SHIFT(keystate) ) 
		{ 
			if (oapiAcceptDelayedKey (OAPI_KEY_E, 1.0))
		{
			

				VESSELSTATUS vs;
				OBJHANDLE hvessel;
				GetStatus (vs);
				vs.eng_main = vs.eng_hovr = 0.0;
				vs.fuel=1.0;
				vs.status=0;
				VECTOR3 ofs = _V(1.8,1.8,24.25);
				Local2Rel (ofs, vs.rpos);
				char name[256], name2[256];
			//	strcpy (name, GetName()); strcat (name, "-PLT");


				UINT i;
				for (i=1;i<8;i++) {
						strcpy (name2, GetName()); 
						sprintf(name, "-EV%d", i);
						strcat (name2, name);
						hvessel=oapiGetVesselByName(name2);
						if (hvessel == 0) break;
				}		
				if ( i < 7) 
				{
					VESSEL::Create (name2, "gemini_leva", vs);
					hvessel=oapiGetVesselByName(name2);
					oapiSetFocusObject(hvessel);
				}

	
			}
			
			UnDockArm = false;
			return 1;
		}
*/
/*		if (KEYDOWN (keystate, OAPI_KEY_D) && KEYMOD_CONTROL(keystate) && !KEYMOD_SHIFT(keystate) ) 
		{ // "Undock"
			if (!UnDockArm)
				UnDockArm = true;
			else {
				Undock(ALLDOCKS);
				UnDockArm = false;
			}
			return 1;
		}
		if (KEYDOWN (keystate, OAPI_KEY_0) && !KEYMOD_CONTROL(keystate) && !KEYMOD_SHIFT(keystate) && UnDockArm ) 
		{ // "Undock the numbered port"
			Undock(0);
			UnDockArm = false;
			return 1;
		}
*/
/*
	return 0;
}
*/
///////////

#define mkrcslin(i,x1,y1,z1,x2,y2,z2,m) {\
	th_att_lin[i] = CreateThruster (_V(x1,y1,z1), _V(x2,y2,z2), cfg->RCSThrust*m, ph_RCS, cfg->IspRCS, cfg->IspRCS);\
	AddExhaust (th_att_lin[i], RCS_EX_LEN*m, RCS_EX_WIDTH*m, 0.357*m,tex);\
}

void ORION::SetAttControlsOrion()
{
	dprint("SetAttControlsOrion");
	//	VECTOR3 v_com; double plmass; plmass = CalculateMass(&v_com); double comshift = -1.0*(v_com.z * plmass)/(GetMass());

	double comshift = 0.0;

	#define RCS_EX_WIDTH 0.15
	#define RCS_EX_LEN 1.5

	THRUSTER_HANDLE th_att_rot[4], th_att_lin[4];

	//CreateThruster (&pos, &dir, maxth0, hp, isp0, isp_ref, p_ref)

	th_att_rot[0] = CreateThruster (_V(0,-1.8, 24.0-comshift), _V(0, 1,0), cfg->GyroThrust, ph_GYRO, cfg->IspGYRO, cfg->IspGYRO);
	th_att_rot[1] = CreateThruster (_V(0, 4.9,-24.0-comshift), _V(0,-1,0), cfg->GyroThrust, ph_GYRO, cfg->IspGYRO, cfg->IspGYRO);
	th_att_rot[2] = CreateThruster (_V(0, 1.8, 24.0-comshift), _V(0,-1,0), cfg->GyroThrust, ph_GYRO, cfg->IspGYRO, cfg->IspGYRO);
	th_att_rot[3] = CreateThruster (_V(0,-4.9,-24.0-comshift), _V(0, 1,0), cfg->GyroThrust, ph_GYRO, cfg->IspGYRO, cfg->IspGYRO);
	CreateThrusterGroup (th_att_rot,   2, THGROUP_ATT_PITCHUP);
	CreateThrusterGroup (th_att_rot+2, 2, THGROUP_ATT_PITCHDOWN);

	th_att_rot[0] = CreateThruster (_V( 1.8,0, 24.0-comshift), _V(-1,0,0), cfg->GyroThrust, ph_GYRO, cfg->IspGYRO, cfg->IspGYRO);
	th_att_rot[1] = CreateThruster (_V(-4.9,0,-24.0-comshift), _V( 1,0,0), cfg->GyroThrust, ph_GYRO, cfg->IspGYRO, cfg->IspGYRO);
	th_att_rot[2] = CreateThruster (_V(-1.8,0, 24.0-comshift), _V( 1,0,0), cfg->GyroThrust, ph_GYRO, cfg->IspGYRO, cfg->IspGYRO);
	th_att_rot[3] = CreateThruster (_V( 4.9,0,-24.0-comshift), _V(-1,0,0), cfg->GyroThrust, ph_GYRO, cfg->IspGYRO, cfg->IspGYRO);
	CreateThrusterGroup (th_att_rot,   2, THGROUP_ATT_YAWLEFT);
	CreateThrusterGroup (th_att_rot+2, 2, THGROUP_ATT_YAWRIGHT);

	th_att_rot[0] = CreateThruster (_V( 4.8,0,0), _V(0, 1,0), cfg->GyroThrust, ph_GYRO, cfg->IspGYRO, cfg->IspGYRO);
	th_att_rot[1] = CreateThruster (_V(-4.8,0,0), _V(0,-1,0), cfg->GyroThrust, ph_GYRO, cfg->IspGYRO, cfg->IspGYRO);
	th_att_rot[2] = CreateThruster (_V(-4.8,0,0), _V(0, 1,0), cfg->GyroThrust, ph_GYRO, cfg->IspGYRO, cfg->IspGYRO);
	th_att_rot[3] = CreateThruster (_V( 4.8,0,0), _V(0,-1,0), cfg->GyroThrust, ph_GYRO, cfg->IspGYRO, cfg->IspGYRO);
	CreateThrusterGroup (th_att_rot,   2, THGROUP_ATT_BANKLEFT);
	CreateThrusterGroup (th_att_rot+2, 2, THGROUP_ATT_BANKRIGHT);


	mkrcslin(0, 3.8,-3.8,30.0, 0,1,0, 1);
	mkrcslin(1, -3.8,-3.8,30.0, 0,1,0, 1);
	mkrcslin(2, 7,-7,-7.5, 0,1,0, 4);
	mkrcslin(3, -7,-7,-7.5, 0,1,0, 4);
	CreateThrusterGroup (th_att_lin,   4, THGROUP_ATT_UP);

	mkrcslin(0, 3.8,3.8,30.0, 0,-1,0, 1);
	mkrcslin(1, -3.8,3.8,30.0, 0,-1,0, 1);
	mkrcslin(2, 7,7,-7.5, 0,-1,0, 4);
	mkrcslin(3, -7,7,-7.5, 0,-1,0, 4);
	CreateThrusterGroup (th_att_lin,   4, THGROUP_ATT_DOWN);

	mkrcslin(0, 3.8,3.8, 30.0, -1,0,0, 1);
	mkrcslin(1, 3.8,-3.8, 30.0, -1,0,0, 1);
	mkrcslin(2, 7,7,-7.5, -1,0,0, 4);
	mkrcslin(3, 7,-7,-7.5, -1,0,0, 4);
	CreateThrusterGroup (th_att_lin,   4, THGROUP_ATT_LEFT);

	mkrcslin(0, -3.8,3.8, 30.0, 1,0,0, 1);
	mkrcslin(1, -3.8,-3.8, 30.0, 1,0,0, 1);
	mkrcslin(2, -7,7,-7.5, 1,0,0, 4);
	mkrcslin(3, -7,-7,-7.5, 1,0,0, 4);
	CreateThrusterGroup (th_att_lin,   4, THGROUP_ATT_RIGHT);

	mkrcslin(0, 7,-7,-7.5, 0,0,1, 2);
	mkrcslin(1, 7,7,-7.5, 0,0,1, 2);
	mkrcslin(2, -7,-7,-7.5, 0,0,1, 2);
	mkrcslin(3, -7,7,-7.5, 0,0,1, 2);
	CreateThrusterGroup (th_att_lin,   4, THGROUP_ATT_FORWARD);

	mkrcslin(0, 3.8,-3.8,30.0, 0,0,-1, 1);
	mkrcslin(1, 3.8,3.8,30.0, 0,0,-1, 1);
	mkrcslin(2, -3.8,-3.8,30.0, 0,0,-1, 1);
	mkrcslin(3, -3.8,3.8,30.0, 0,0,-1, 1);
	CreateThrusterGroup (th_att_lin, 4, THGROUP_ATT_BACK);

}


////////////
/*
#define RCS2 1

void ORION::SetAttControlsOrion()
{
dprint("SetAttControlsOrion");
//	VECTOR3 v_com;
//	double plmass;

	//plmass = CalculateMass(&v_com);
/////
//	double comshift = -1.0*(v_com.z * plmass)/(GetMass());

	double comshift = 0.0;
//#define RCS_EX_WIDTH 0.1
//#define RCS_EX_LEN 1.5

#define RCS_EX_WIDTH 0.9
#define RCS_EX_LEN 9.0

#define RCSBXY 9.5
#define RCSBZ 7.7

#define RCSFXY 5
#define RCSFZ 30

	THRUSTER_HANDLE th_att_rot[4], th_att_lin[4];

	//CreateThruster (&pos, &dir, maxth0, hp, isp0, isp_ref, p_ref)

	th_att_rot[0] = CreateThruster (_V(0,-1.8, 24.0-comshift), _V(0, 1,0), cfg->RCSThrust, ph_RCS, cfg->IspGYRO, cfg->IspRCS);
	th_att_rot[1] = CreateThruster (_V(0, 4.9,-24.0-comshift), _V(0,-1,0), cfg->RCSThrust, ph_RCS, cfg->IspGYRO, cfg->IspRCS);
	th_att_rot[2] = CreateThruster (_V(0, 1.8, 24.0-comshift), _V(0,-1,0), cfg->RCSThrust, ph_RCS, cfg->IspGYRO, cfg->IspRCS);
	th_att_rot[3] = CreateThruster (_V(0,-4.9,-24.0-comshift), _V(0, 1,0), cfg->RCSThrust, ph_RCS, cfg->IspGYRO, cfg->IspRCS);
	CreateThrusterGroup (th_att_rot,   2, THGROUP_ATT_PITCHUP);
	CreateThrusterGroup (th_att_rot+2, 2, THGROUP_ATT_PITCHDOWN);
//	AddExhaust (th_att_rot[0], RCS_EX_LEN, RCS_EX_WIDTH, tex);
//	AddExhaust (th_att_rot[1], RCS_EX_LEN, RCS_EX_WIDTH, tex);
//	AddExhaust (th_att_rot[2], RCS_EX_LEN, RCS_EX_WIDTH, tex);
//	AddExhaust (th_att_rot[3], RCS_EX_LEN, RCS_EX_WIDTH, tex);
#ifdef RCS2 
	th_att_lin[0] = CreateThruster (_V(0,-5.0, 30.0), _V(0, 1,0), cfg->RCSThrust, ph_RCS, cfg->IspRCS, cfg->IspRCS);
	th_att_lin[3] = CreateThruster (_V(0, 9.5,-7.5), _V(0,-1,0), cfg->RCSThrust*4, ph_RCS, cfg->IspRCS, cfg->IspRCS);
	th_att_lin[2] = CreateThruster (_V(0, 5.0, 30.0), _V(0,-1,0), cfg->RCSThrust, ph_RCS, cfg->IspRCS, cfg->IspRCS);
	th_att_lin[1] = CreateThruster (_V(0,-9.5,-7.5), _V(0, 1,0), cfg->RCSThrust*4, ph_RCS, cfg->IspRCS, cfg->IspRCS);
#else
	th_att_lin[0] = CreateThruster (_V(0,-1.8, 24.0), _V(0, 1,0), cfg->RCSThrust, ph_RCS, cfg->IspRCS, cfg->IspRCS);
	th_att_lin[3] = CreateThruster (_V(0, 4.9,-6.0), _V(0,-1,0), cfg->RCSThrust*4, ph_RCS, cfg->IspRCS, cfg->IspRCS);
	th_att_lin[2] = CreateThruster (_V(0, 1.8, 24.0), _V(0,-1,0), cfg->RCSThrust, ph_RCS, cfg->IspRCS, cfg->IspRCS);
	th_att_lin[1] = CreateThruster (_V(0,-4.9,-6.0), _V(0, 1,0), cfg->RCSThrust*4, ph_RCS, cfg->IspRCS, cfg->IspRCS);
#endif
	CreateThrusterGroup (th_att_lin,   2, THGROUP_ATT_UP);
	CreateThrusterGroup (th_att_lin+2, 2, THGROUP_ATT_DOWN);
	AddExhaust (th_att_lin[0], RCS_EX_LEN, RCS_EX_WIDTH, tex);
	AddExhaust (th_att_lin[1], RCS_EX_LEN, RCS_EX_WIDTH, tex);
	AddExhaust (th_att_lin[2], RCS_EX_LEN, RCS_EX_WIDTH, tex);
	AddExhaust (th_att_lin[3], RCS_EX_LEN, RCS_EX_WIDTH, tex);

	th_att_rot[0] = CreateThruster (_V( 1.8,0, 24.0-comshift), _V(-1,0,0), cfg->RCSThrust, ph_RCS, cfg->IspGYRO, cfg->IspRCS);
	th_att_rot[1] = CreateThruster (_V(-4.9,0,-24.0-comshift), _V( 1,0,0), cfg->RCSThrust, ph_RCS, cfg->IspGYRO, cfg->IspRCS);
	th_att_rot[2] = CreateThruster (_V(-1.8,0, 24.0-comshift), _V( 1,0,0), cfg->RCSThrust, ph_RCS, cfg->IspGYRO, cfg->IspRCS);
	th_att_rot[3] = CreateThruster (_V( 4.9,0,-24.0-comshift), _V(-1,0,0), cfg->RCSThrust, ph_RCS, cfg->IspGYRO, cfg->IspRCS);
	CreateThrusterGroup (th_att_rot,   2, THGROUP_ATT_YAWLEFT);
	CreateThrusterGroup (th_att_rot+2, 2, THGROUP_ATT_YAWRIGHT);
//	AddExhaust (th_att_rot[0], RCS_EX_LEN, RCS_EX_WIDTH, tex);
//	AddExhaust (th_att_rot[1], RCS_EX_LEN, RCS_EX_WIDTH, tex);
//	AddExhaust (th_att_rot[2], RCS_EX_LEN, RCS_EX_WIDTH, tex);
//	AddExhaust (th_att_rot[3], RCS_EX_LEN, RCS_EX_WIDTH, tex);
#ifdef RCS2
	th_att_lin[0] = CreateThruster (_V( 5.0,0, 30.0), _V(-1,0,0), cfg->RCSThrust, ph_RCS, cfg->IspRCS, cfg->IspRCS);
	th_att_lin[3] = CreateThruster (_V(-9.5,0,-7.5), _V( 1,0,0), cfg->RCSThrust*4, ph_RCS, cfg->IspRCS, cfg->IspRCS);
	th_att_lin[2] = CreateThruster (_V(-5.0,0, 30.0), _V( 1,0,0), cfg->RCSThrust, ph_RCS, cfg->IspRCS, cfg->IspRCS);
	th_att_lin[1] = CreateThruster (_V( 9.5,0,-7.5), _V(-1,0,0), cfg->RCSThrust*4, ph_RCS, cfg->IspRCS, cfg->IspRCS);
#else
	th_att_lin[0] = CreateThruster (_V( 1.9,0, 24.0), _V(-1,0,0), cfg->RCSThrust, ph_RCS, cfg->IspRCS, cfg->IspRCS);
	th_att_lin[3] = CreateThruster (_V(-4.9,0,-6.0), _V( 1,0,0), cfg->RCSThrust*4, ph_RCS, cfg->IspRCS, cfg->IspRCS);
	th_att_lin[2] = CreateThruster (_V(-1.9,0, 24.0), _V( 1,0,0), cfg->RCSThrust, ph_RCS, cfg->IspRCS, cfg->IspRCS);
	th_att_lin[1] = CreateThruster (_V( 4.9,0,-6.0), _V(-1,0,0), cfg->RCSThrust*4, ph_RCS, cfg->IspRCS, cfg->IspRCS);
#endif
	CreateThrusterGroup (th_att_lin,   2, THGROUP_ATT_LEFT);
	CreateThrusterGroup (th_att_lin+2, 2, THGROUP_ATT_RIGHT);
	AddExhaust (th_att_lin[0], RCS_EX_LEN, RCS_EX_WIDTH, tex);
	AddExhaust (th_att_lin[1], RCS_EX_LEN, RCS_EX_WIDTH, tex);
	AddExhaust (th_att_lin[2], RCS_EX_LEN, RCS_EX_WIDTH, tex);
	AddExhaust (th_att_lin[3], RCS_EX_LEN, RCS_EX_WIDTH, tex);

	th_att_rot[0] = CreateThruster (_V( 4.8,0,0), _V(0, 1,0), cfg->RCSThrust, ph_RCS, cfg->IspGYRO, cfg->IspRCS);
	th_att_rot[1] = CreateThruster (_V(-4.8,0,0), _V(0,-1,0), cfg->RCSThrust, ph_RCS, cfg->IspGYRO, cfg->IspRCS);
	th_att_rot[2] = CreateThruster (_V(-4.8,0,0), _V(0, 1,0), cfg->RCSThrust, ph_RCS, cfg->IspGYRO, cfg->IspRCS);
	th_att_rot[3] = CreateThruster (_V( 4.8,0,0), _V(0,-1,0), cfg->RCSThrust, ph_RCS, cfg->IspGYRO, cfg->IspRCS);
	CreateThrusterGroup (th_att_rot,   2, THGROUP_ATT_BANKLEFT);
	CreateThrusterGroup (th_att_rot+2, 2, THGROUP_ATT_BANKRIGHT);
//	AddExhaust (th_att_rot[0], RCS_EX_LEN, RCS_EX_WIDTH, tex);
//	AddExhaust (th_att_rot[1], RCS_EX_LEN, RCS_EX_WIDTH, tex);
//	AddExhaust (th_att_rot[2], RCS_EX_LEN, RCS_EX_WIDTH, tex);
//	AddExhaust (th_att_rot[3], RCS_EX_LEN, RCS_EX_WIDTH, tex);
#ifdef RCS2
	th_att_lin[0] = CreateThruster (_V( 9.5,0, -7.5), _V(0,0, 1), cfg->RCSThrust*4, ph_RCS, cfg->IspRCS, cfg->IspRCS);
	th_att_lin[1] = CreateThruster (_V(-9.5,0, -7.5), _V(0,0, 1), cfg->RCSThrust*4, ph_RCS, cfg->IspRCS, cfg->IspRCS);
	th_att_lin[2] = CreateThruster (_V(-5.0,0, 30.0), _V(0,0,-1), cfg->RCSThrust, ph_RCS, cfg->IspRCS, cfg->IspRCS);
	th_att_lin[3] = CreateThruster (_V( 5.0,0, 30.0), _V(0,0,-1), cfg->RCSThrust, ph_RCS, cfg->IspRCS, cfg->IspRCS);
#else
	th_att_lin[0] = CreateThruster (_V( 4.9,0, -6.0), _V(0,0, 1), cfg->RCSThrust*4, ph_RCS, cfg->IspRCS, cfg->IspRCS);
	th_att_lin[1] = CreateThruster (_V(-4.9,0, -6.0), _V(0,0, 1), cfg->RCSThrust*4, ph_RCS, cfg->IspRCS, cfg->IspRCS);
	th_att_lin[2] = CreateThruster (_V(-1.9,0, 24.0), _V(0,0,-1), cfg->RCSThrust*4, ph_RCS, cfg->IspRCS, cfg->IspRCS);
	th_att_lin[3] = CreateThruster (_V( 1.9,0, 24.0), _V(0,0,-1), cfg->RCSThrust*4, ph_RCS, cfg->IspRCS, cfg->IspRCS);
#endif
	CreateThrusterGroup (th_att_lin,   2, THGROUP_ATT_FORWARD);
	CreateThrusterGroup (th_att_lin+2, 2, THGROUP_ATT_BACK);
	AddExhaust (th_att_lin[0], RCS_EX_LEN, RCS_EX_WIDTH, tex);
	AddExhaust (th_att_lin[1], RCS_EX_LEN, RCS_EX_WIDTH, tex);
	AddExhaust (th_att_lin[2], RCS_EX_LEN, RCS_EX_WIDTH, tex);
	AddExhaust (th_att_lin[3], RCS_EX_LEN, RCS_EX_WIDTH, tex);
}
*/

////////////
/*
void ORION::clbkDrawHUD(int mode, const HUDPAINTSPEC *hps, HDC hDC) {
	VESSEL2::clbkDrawHUD (mode, hps, hDC);
	side_offset + 10, _from_top
*/

///////////////////////////// cogspew
#define shiftlingr(g,str){int thcount=GetGroupThrusterCount(g);\
	for (int i = 0; i < thcount; i++) {\
		th = GetGroupThruster(g, i);\
		GetThrusterRef(th, thref);\
		thmax=GetThrusterMax0(th);\
		dprint("TH @ %.2f,%.2f,%.2f: %.4f",thref,thmax);\
		if(thref.z==RCS_LIN_F_ZPOS) {\
			SetThrusterMax0(th,cfg->RCSThrust*(fzratio/1));\
			dprint("                `   ratio %.4f",fzratio);\
		} else if (thref.z==RCS_LIN_B_ZPOS) {\
			SetThrusterMax0(th,cfg->RCSThrust*(bzratio/1));\
			dprint("                `   ratio %.4f",bzratio);\
		}\
		thmax=GetThrusterMax0(th);\
		dprint("                ---> %.4f",thmax);\
	}\
}

void ORION::rcs_linshift(int onoff) {
	VECTOR3 thref;
	THRUSTER_HANDLE th;
	double thmax;
	GetSuperstructureCG(shipcg);
	double bzratio=RCS_LIN_B_RATIO;
	double fzratio=RCS_LIN_F_RATIO;

	/*	HURR-DURR: calc via sqrt ( ((cg.x+ref.x)^2)+((cg.z+ref.z)^2) ) etc */
	if(onoff>0) {
		double fzdiff=RCS_LIN_F_ZPOS-(shipcg.z);
		double bzdiff=(RCS_LIN_B_ZPOS-(RCS_LIN_B_ZPOS*2))+(shipcg.z);
		if(fzdiff <= 0 || bzdiff <= 0) {
			if (onoff==1) syslog("shift-z: cg: %.8f, out of bounds.",shipcg.z);
		} else {
			bzratio=fzdiff/bzdiff;									// 1.14
			if(bzratio>RCS_LIN_B_RATIO) {							// 1.14>4
				fzratio=RCS_LIN_B_RATIO/bzratio;					// 
				bzratio=RCS_LIN_B_RATIO;							//
			}
		}
		if (onoff==1) syslog("shift-z: zratio f/b: %.8f/%.8f",fzratio,bzratio);
	} else {
		syslog("shift-z: off. zratio f/b: %.8f/%.8f",fzratio,bzratio);		
	}
	shiftlingr(THGROUP_ATT_UP,"U");
	shiftlingr(THGROUP_ATT_DOWN,"D");
	shiftlingr(THGROUP_ATT_LEFT,"L");
	shiftlingr(THGROUP_ATT_RIGHT,"R");
	return;
}


////////////
// Initialisation
DLLCLBK VESSEL *ovcInit (OBJHANDLE hvessel, int flightmodel)
{
	ORION *Orion= new ORION( hvessel, flightmodel );
	return Orion;
}

// Cleanup
DLLCLBK void ovcExit (VESSEL *vessel)
{
	if (vessel) delete (ORION*)vessel;
}
